# frozen_string_literal: true

module Toys
  ##
  # The core Toys classes.
  #
  module Core
    ##
    # Current version of Toys core.
    # @return [String]
    #
    VERSION = "0.21.0"
  end

  ##
  # Deprecated
  #
  # @private
  #
  CORE_VERSION = Core::VERSION
end
